#include "SamplingBehavior.h"
#include "Model.h"


EtaSamplingBehavior::EtaSamplingBehavior(double lower_bound, double upper_bound) : SamplingBehavior(lower_bound, upper_bound) {}

vector<int> EtaSamplingBehavior::component_nodes(vector<int> component, int component_index, int n_g) {
	vector<int> result;
	for (int i = n_g; i < component.size(); i++) {
		if (component[i] == component_index) {
			result.push_back(i);
		}
	}

	return result;
}

vector<int> EtaSamplingBehavior::component_generators(vector<int> component, int component_index, int n_g) {
	vector<int> result;
	for (int i = 0; i < n_g; i++) {
		if (component[i] == component_index) {
			result.push_back(i);
		}
	}

	return result;
}

void EtaSamplingBehavior::balance_power(Angle_Eta &model, int num, vector<int> component) {
	//ensure that each cluster has one and only one reference generator
	for (int i = 0; i < num; i++) {
		bool has_reference = false;

		vector<int> generators = component_generators(component, i, model.n_g);
		for (const auto &generator : generators) {
			if (find(model.reference_generators.begin(), model.reference_generators.end(), generator) != model.reference_generators.end()) {
				has_reference = true;
				break;
			}
		}
		if (!has_reference) {
			model.reference_generators.push_back(generators[0]); //Fails if no generators in cluster
			
			//Shift angles
			for(int j = 0; j < generators.size(); j++) {
				model.angles[generators[j]] -= model.angles[generators[0]];
			}
			vector<int> nodes = component_nodes(component, i, model.n_g);
			for(int j = 0; j < nodes.size(); j++) {
				model.angles[nodes[j]] -= model.angles[generators[0]];	
			}
		}
	}

	//carry out necessary load shedding and/or generation cutting
	vector<double> power;
	for (int i = 0; i < num; i++) {
		power.push_back(0.);
	}

	//compute power in each cluster
	for (int i = 0; i < component.size(); i++) {
		power[component[i]] += model.P[i];
	}

	//rebalance power
	for (int i = 0; i < num; i++) { //loop over clusters
		if (power[i] > 0) { // not enough production . load shedding
			vector<int> nodes = component_nodes(component, i, model.n_g);

			double demand_surplus_at_node;

			for (int j = 0; j < nodes.size(); j++) {
				demand_surplus_at_node = min(power[i]/(nodes.size() - j), model.P[nodes[j]]);

				model.P[nodes[j]] -= demand_surplus_at_node;
				power[i] -= demand_surplus_at_node;
			}
			// for (int j = 0; j < nodes.size(); j++) {
			// 	model.P[nodes[j]] -= power[i]/nodes.size();
			// }

		}
		else if (power[i] < 0) { // too much production . production cutting
			vector<int> generators = component_generators(component, i, model.n_g);

			double production_surplus_at_generator;

			for (int j = 0; j < generators.size(); j++) {
				production_surplus_at_generator = max(power[i]/(generators.size() - j), model.P[generators[j]]);

				model.P[generators[j]] -= production_surplus_at_generator;
				power[i] -= production_surplus_at_generator;
			}
			// for (int j = 0; j < generators.size(); j++) {
			// 	montecarlo.P[generators[j]] -= power[i]/generators.size();
			// }
		}
	}
	// cout << "Power = ";
	// for (auto &p : montecarlo.P) {
	// 	cout << p << " ";
	// }
	// cout << endl;

	// cout << "reference generators = ";
	// for (auto &gen : montecarlo.reference_generators) {
	// 	cout << gen << " ";
	// }
	// cout << endl;
}

void EtaSamplingBehavior::manage_lines(Angle_Eta &model, double new_value, double old_value, int index) {
	bool edge_changed = false;
	int node_1;
	int node_2;
	double threshold = 0.2;

	if (new_value < threshold && old_value >= threshold) {
		node_1 = model.index_to_lines[index][0];
		node_2 = model.index_to_lines[index][1];
		remove_edge(node_1, node_2, model.G);

		edge_changed = true;
	}
	else if (new_value >= threshold && old_value < threshold) {
		node_1 = model.index_to_lines[index][0];
		node_2 = model.index_to_lines[index][1];
		add_edge(node_1, node_2, model.G);

		edge_changed = true;
	}
	if (edge_changed) {
		//load shedding or generation cutting

		//calculate new connected components
		vector<int> component(num_vertices(model.G));
    	int num = connected_components(model.G, &component[0]);

		// cout << "\nEdge changed with new number of components = " << num << " new_value = " << new_value << " old_value = " << old_value << "node_1 = " << node_1 << "node_2 = " << node_2 << "\n";
		// for (int i = 0; i < component.size(); i++) {
		// 	cout << component[i] << " ";
		// }
		// cout << endl;
		// cout << endl;

		//check if changes to connected components
		if (num != model.n_components) {
			model.P = model.base_P;
			model.reference_generators.erase(model.reference_generators.begin()+1, model.reference_generators.end());
			
			balance_power(model, num, component);

			model.n_components = num;
		}
	}
}